/** 
 * Copyright (C) 2013 Quan Wang <wangq10@rpi.edu>
 * Signal Analysis and Machine Perception Laboratory
 * Department of Electrical, Computer, and Systems Engineering
 * Rensselaer Polytechnic Institute, Troy, NY 12180, USA
 */

import java.util.Vector;
import Jama.*;

/**
 * This class is for the training of binary Fisher LDA
 * 
 * @author Quan Wang
 *
 */

public class FisherTrain {

	// training
	public static FisherLdaModel train(double[][] features, int[] labels)
	{
		if(features.length!=labels.length)
		{
			System.out.println("Error: feature size does not match label length!");
			return null;
		}

		Vector<Integer> ind0=new Vector<Integer>();
		Vector<Integer> ind1=new Vector<Integer>();

		// indices for label 0 and label 1
		for(int i=0;i<labels.length;i++)
		{
			if(labels[i]==0)
			{
				ind0.add(i);
			}
			else if(labels[i]==1)
			{
				ind1.add(i);
			}
			else
			{
				System.out.println("Error: labels can only contain values of 0 or 1!");
				return null;
			}
		}

		double[] mu0=new double[features[0].length];
		double[] mu1=new double[features[0].length];
		Matrix S0=new Matrix(features[0].length,features[0].length);
		Matrix S1=new Matrix(features[0].length,features[0].length);

		// compute mu0 and mu1
		for(int i=0;i<ind0.size();i++)
		{
			for(int j=0;j<features[0].length;j++)
			{
				mu0[j]+=features[ind0.get(i)][j];
			}
		}
		for(int i=0;i<ind1.size();i++)
		{
			for(int j=0;j<features[0].length;j++)
			{
				mu1[j]+=features[ind1.get(i)][j];
			}
		}
		for(int j=0;j<features[0].length;j++)
		{
			mu0[j]/=(double)ind0.size();
			mu1[j]/=(double)ind1.size();
		}
		
		// compute S0 and S1 (covariance matrices)
		for(int i=0;i<ind0.size();i++)
		{
			double[] row=new double[features[0].length];
			for(int j=0;j<row.length;j++)
			{
				row[j]=features[ind0.get(i)][j]-mu0[j];
			}
			Matrix rowVector=new Matrix(row,row.length);
			S0.plusEquals(rowVector.times(rowVector.transpose()));
		}
		S0.timesEquals(1/(double)ind0.size());
		for(int i=0;i<ind1.size();i++)
		{
			double[] row=new double[features[0].length];
			for(int j=0;j<row.length;j++)
			{
				row[j]=features[ind1.get(i)][j]-mu1[j];
			}
			Matrix rowVector=new Matrix(row,row.length);
			S1.plusEquals(rowVector.times(rowVector.transpose()));
		}
		S1.timesEquals(1/(double)ind1.size());
		
		// solve w (weights)
		double[] dmu=new double[mu0.length];
		for(int i=0;i<mu0.length;i++)
		{
			dmu[i]=mu1[i]-mu0[i];
		}
		Matrix dmuVector=new Matrix(dmu,dmu.length);
		Matrix wVector=S0.plus(S1).solve(dmuVector);
		double wNorm=0;
		for(int i=0;i<dmu.length;i++)
		{
			wNorm+=wVector.get(i, 0)*wVector.get(i, 0);
		}
		wNorm=Math.sqrt(wNorm);
		wVector.timesEquals(1/wNorm);

		// get t (threshold)
		Matrix fMatrix=new Matrix(features); // features
		Matrix fpVector=fMatrix.times(wVector); // projected features
		double pmu0=0; // mean
		double pmu1=0; // mean
		double pS0=0; // variance
		double pS1=0; // variance
		
		for(int i=0;i<ind0.size();i++)
		{
			pmu0+=fpVector.get(ind0.get(i), 0);
			
		}
		pmu0/=(double)ind0.size();
		for(int i=0;i<ind1.size();i++)
		{
			pmu1+=fpVector.get(ind1.get(i), 0);
		}
		pmu1/=(double)ind1.size();
		
		for(int i=0;i<ind0.size();i++)
		{
			pS0+=(fpVector.get(ind0.get(i), 0)-pmu0)*(fpVector.get(ind0.get(i), 0)-pmu0);
		}
		pS0/=(double)ind0.size();
		for(int i=0;i<ind1.size();i++)
		{
			pS1+=(fpVector.get(ind1.get(i), 0)-pmu1)*(fpVector.get(ind1.get(i), 0)-pmu1);
		}
		pS1/=(double)ind1.size();
		
		// pa, pb and pc are constants of the quadratic equation, and pD is b2-4ac
		double pa=pS1-pS0;
		double pb=-2*(pmu0*pS1-pmu1*pS0);
		double pc=pS1*pmu0*pmu0-pS0*pmu1*pmu1-2*pS0*pS1*Math.log((double)ind0.size()/(double)ind1.size());
		double pD=pb*pb-4*pa*pc;
		
		if(pD<0)
		{
			System.out.println("Error: no threshold found!");
			return null;
		}
		
		double t1=(-pb+Math.sqrt(pD))/2/pa;
		double t2=(-pb-Math.sqrt(pD))/2/pa;
		double t; // threshold
		if((t1-pmu0)*(t1-pmu0)<(t2-pmu0)*(t2-pmu0))
		{
			t=t1;
		}
		else
		{
			t=t2;
		}
		
		
		FisherLdaModel model=new FisherLdaModel(wVector.getColumnPackedCopy(),t,fpVector.getColumnPackedCopy());

		return model;
	}

}
